import type { Behavior } from "../observation/behavior.js";
/**
 * A node that can be targeted by styles.
 * @public
 */
export interface StyleTarget {
    /**
     * Stylesheets to be adopted by the node.
     */
    adoptedStyleSheets?: CSSStyleSheet[];
    /**
     * Adds styles to the target by appending the styles.
     * @param styles - The styles element to add.
     */
    append(styles: HTMLStyleElement): void;
    /**
     * Adds styles to the target by prepending the styles.
     * @param styles - The styles element to add.
     * @deprecated - use append()
     */
    prepend(styles: HTMLStyleElement): void;
    /**
     * Removes styles from the target.
     * @param styles - The styles element to remove.
     */
    removeChild(styles: HTMLStyleElement): void;
    /**
     * Returns all element descendants of node that match selectors.
     * @param selectors - The CSS selector to use for the query.
     */
    querySelectorAll<E extends Element = Element>(selectors: string): NodeListOf<E>;
}
/**
 * Represents styles that can be composed into the ShadowDOM of a custom element.
 * @public
 */
export declare type ComposableStyles = string | ElementStyles | CSSStyleSheet;
/**
 * Creates an ElementStyles instance for an array of ComposableStyles.
 * @public
 */
export declare type ElementStyleFactory = (styles: ReadonlyArray<ComposableStyles>) => ElementStyles;
/**
 * Represents styles that can be applied to a custom element.
 * @public
 */
export declare abstract class ElementStyles {
    private targets;
    /** @internal */
    abstract readonly styles: ReadonlyArray<ComposableStyles>;
    /** @internal */
    abstract readonly behaviors: ReadonlyArray<Behavior> | null;
    /** @internal */
    addStylesTo(target: StyleTarget): void;
    /** @internal */
    removeStylesFrom(target: StyleTarget): void;
    /** @internal */
    isAttachedTo(target: StyleTarget): boolean;
    /**
     * Associates behaviors with this set of styles.
     * @param behaviors - The behaviors to associate.
     */
    withBehaviors(...behaviors: Behavior[]): this;
    /**
     * Create ElementStyles from ComposableStyles.
     */
    static readonly create: ElementStyleFactory;
}
/**
 * A Symbol that can be added to a CSSStyleSheet to cause it to be prepended (rather than appended) to adoptedStyleSheets.
 * @public
 */
export declare const prependToAdoptedStyleSheetsSymbol: unique symbol;
/**
 * https://wicg.github.io/construct-stylesheets/
 * https://developers.google.com/web/updates/2019/02/constructable-stylesheets
 *
 * @internal
 */
export declare class AdoptedStyleSheetsStyles extends ElementStyles {
    styles: ComposableStyles[];
    private styleSheetCache;
    private _styleSheets;
    private get styleSheets();
    readonly behaviors: ReadonlyArray<Behavior> | null;
    constructor(styles: ComposableStyles[], styleSheetCache: Map<string, CSSStyleSheet>);
    addStylesTo(target: Required<StyleTarget>): void;
    removeStylesFrom(target: Required<StyleTarget>): void;
}
/**
 * @internal
 */
export declare class StyleElementStyles extends ElementStyles {
    styles: ComposableStyles[];
    private readonly styleSheets;
    private readonly styleClass;
    readonly behaviors: ReadonlyArray<Behavior> | null;
    constructor(styles: ComposableStyles[]);
    addStylesTo(target: StyleTarget): void;
    removeStylesFrom(target: StyleTarget): void;
    isAttachedTo(target: StyleTarget): boolean;
    private normalizeTarget;
}
